(*
Author:  Akihisa Yamada (2016-2017)
License: LGPL (see file COPYING.LESSER)
*)
theory Max_Polynomial
  imports Ordered_Algebra
    "../LTS/Integer_Arithmetic"
    Polynomial_Interpolation.Ring_Hom
    "HOL-Algebra.Ring"
begin

section {* Max-Polynomial Interpretations *}

subsection {* Missing product_lists *}

lemma product_lists_eq_Nil_iff [simp]: "product_lists xss = [] \<longleftrightarrow> [] \<in> set xss"
  by (induct xss, auto)
text {*
  Clearly the above simp rule is in a good direction. But also
  @{term "product_lists xss = []"} is a good simplification rule.
*}

lemma product_lists_Nil[simp]: "[] \<in> set xss \<Longrightarrow> product_lists xss = []" by simp

primrec product_indices where
  "product_indices [] = [[]]"
| "product_indices (n#ns) = concat (map (\<lambda>x. map (Cons x) (product_indices ns)) [0..<n])"

lemma length_product_indices[simp]:
  "length (product_indices ns) = prod_list ns"
  by (induct ns, auto simp: length_concat o_def sum_list_triv)

lemma product_lists_indices:
  "product_lists xss = map (zipf nth xss) (product_indices (map length xss))"
  by (induct xss, auto simp: o_def map_concat intro!: arg_cong[of _ _ concat] nth_equalityI)

lemma length_product_lists' [simp]:
  "length (product_lists xss) = prod_list (map length xss)"
  by (simp add: product_lists_indices)

lemma product_lists_map_map:
  "product_lists (map (map f) xss) = map (map f) (product_lists xss)"
  by (induct xss, auto simp: o_def product_lists_indices)

lemma product_lists_elements:
  "[] \<notin> set xss \<Longrightarrow> \<Union> set (map set (product_lists xss)) = \<Union> set (map set xss)"
  by (induct xss, auto)

lemma product_lists_elements2:
  assumes "xs \<in> set (product_lists xss)" and "x \<in> set xs"
  shows "\<exists>xs' \<in> set xss. x \<in> set xs'"
  by (insert assms product_lists_elements, cases "[] \<in> set xss", auto)

lemma in_set_product_lists_nth:
  assumes "ys \<in> set (product_lists xss)"
    and "i < length xss"
  shows "ys ! i \<in> set (xss ! i)"
proof (insert assms, induct xss arbitrary: i ys)
  case (Cons xs xss)
  then show ?case by (cases i, auto)
qed simp


subsection {* Fundamental Results *}

lemma prod_list_map_filter: (* MOVE-- a counterpart of sum_list_map_filter *)
  assumes "\<And>x. x \<in> set xs \<Longrightarrow> \<not> P x \<Longrightarrow> f x = 1"
  shows "prod_list (map f (filter P xs)) = prod_list (map f xs)"
  using assms by (induct xs, auto)

(*TODO: in SN_Orders -- assumptions could be stated nicer *)
context non_strict_order
begin
lemma antisym: "x \<le> y \<Longrightarrow> y \<le> x \<Longrightarrow> x = y"
  using max_comm[of x y] by (auto simp: max_def)
lemma total: "x \<le> y \<or> y \<le> x"
  by (metis max_ge_y max_def)
end

(* TODO: Names conflict *)
hide_class SN_Orders.ordered_semiring_0 SN_Orders.ordered_semiring_1

(* TODO: Propose for distribution *)

context zero_less_one begin

subclass zero_neq_one using zero_less_one by (unfold_locales, blast)

lemma zero_le_one[intro!]: "0 \<le> 1" by (auto simp: le_less)

end

class ordered_semigroup_mult_zero = order + semigroup_mult + mult_zero +
  assumes mult_right_mono: "\<And>a b c. a \<le> b \<Longrightarrow> c \<ge> 0 \<Longrightarrow> a * c \<le> b * c"
  assumes mult_left_mono: "\<And>a b c. a \<le> b \<Longrightarrow> c \<ge> 0 \<Longrightarrow> c * a \<le> c * b"
begin

lemma mult_mono: "a \<le> b \<Longrightarrow> c \<le> d \<Longrightarrow> 0 \<le> b \<Longrightarrow> 0 \<le> c \<Longrightarrow> a * c \<le> b * d"
  apply (erule (1) mult_right_mono [THEN order_trans])
  apply (erule (1) mult_left_mono)
  done

lemma mult_mono': "a \<le> b \<Longrightarrow> c \<le> d \<Longrightarrow> 0 \<le> a \<Longrightarrow> 0 \<le> c \<Longrightarrow> a * c \<le> b * d"
  by (rule mult_mono) (fast intro: order_trans)+

lemma mult_nonneg_nonneg [simp]: "0 \<le> a \<Longrightarrow> 0 \<le> b \<Longrightarrow> 0 \<le> a * b"
  using mult_left_mono [of 0 b a] by simp

lemma mult_nonneg_nonpos: "0 \<le> a \<Longrightarrow> b \<le> 0 \<Longrightarrow> a * b \<le> 0"
  using mult_left_mono [of b 0 a] by simp

lemma mult_nonpos_nonneg: "a \<le> 0 \<Longrightarrow> 0 \<le> b \<Longrightarrow> a * b \<le> 0"
  using mult_right_mono [of a 0 b] by simp

text \<open>Legacy -- use @{thm [source] mult_nonpos_nonneg}.\<close>
lemma mult_nonneg_nonpos2: "0 \<le> a \<Longrightarrow> b \<le> 0 \<Longrightarrow> b * a \<le> 0"
  by (drule mult_right_mono [of b 0]) auto

lemma split_mult_neg_le: "(0 \<le> a \<and> b \<le> 0) \<or> (a \<le> 0 \<and> 0 \<le> b) \<Longrightarrow> a * b \<le> 0"
  by (auto simp add: mult_nonneg_nonpos mult_nonneg_nonpos2)

end

subclass (in Rings.ordered_semiring_0) ordered_semigroup_mult_zero
  by (unfold_locales; fact mult_right_mono mult_left_mono)

class ordered_monoid_mult_zero = ordered_semigroup_mult_zero + monoid_mult + zero_less_one
begin

lemma prod_list_nonneg:
  assumes "\<And>x. x \<in> set xs \<Longrightarrow> x \<ge> 0"
  shows "prod_list xs \<ge> 0"
  using assms by (induct xs, auto)

end


class ordered_semiring_1 = ordered_semiring_0 + monoid_mult + zero_less_one
begin

subclass semiring_1..

lemma prod_list_nonneg:
  assumes "\<And>x. x \<in> set xs \<Longrightarrow> x \<ge> 0"
  shows "prod_list xs \<ge> 0"
  using assms by (induct xs, auto)

lemma prod_list_nonneg_nth:
  assumes "\<And>i. i < length xs \<Longrightarrow> xs ! i \<ge> 0"
  shows "prod_list xs \<ge> 0"
  using assms by (auto intro!: prod_list_nonneg simp: in_set_conv_nth)

end

text {* Currently (Isabelle 2017), @{class linordered_semiring_1} is not a subclass of
  @{class ordered_semiring_1}, but in the difference everything is negative, so the axioms of
  @{class ordered_semiring} make no sense.
 *}
lemma (in linordered_semiring_1)
  assumes a0: "0 < a" shows "0 < 1"
proof (rule ccontr)
  assume "\<not> 0 < 1" then have "1 \<le> 0" by auto
  from mult_nonpos_nonneg[OF this] a0 have "a \<le> 0" by auto
  with a0 show False by auto
qed

class linordered_semiring_1 = linordered_semiring + semiring_1 + zero_less_one

subclass (in linordered_semiring_1) ordered_semiring_1..

instance nat :: linordered_semiring_1..
instance int :: linordered_semiring_1..
instance rat :: linordered_semiring_1..
instance real :: linordered_semiring_1..


(*TODO: Move! *)
lemma concat_map_map [simp]: "concat (map (map f) xss) = map f (concat xss)"
  by (induct xss, auto)

(* TODO: Move *)
lemma Max_le_Max:
  assumes X: "finite X" "X \<noteq> {}" and Y: "finite Y" "Y \<noteq> {}"
  shows "Max X \<le> Max Y \<longleftrightarrow> (\<forall>x\<in>X. \<exists>y\<in>Y. x \<le> y)"
  by (meson Max.boundedE Max_ge_iff Max_in assms)

(* TODO: Move *)
lemma Max_less_Max:
  assumes X: "finite X" "X \<noteq> {}" and Y: "finite Y" "Y \<noteq> {}"
  shows "Max X < Max Y \<longleftrightarrow> (\<forall>x\<in>X. \<exists>y\<in>Y. x < y)"
  by (meson Max_less_iff assms eq_Max_iff less_le_trans)

definition max_list1 where
  "max_list1 xs \<equiv> if xs = [] then undefined else foldr max (butlast xs) (last xs)"

lemma max_list1_Cons[simp]: "max_list1 (x#xs) = (if xs = [] then x else max x (max_list1 xs))"
  by (auto simp: max_list1_def)

text {* @{const max_list1} is @{term "Max \<circ> set"}, but the latter require a total order. *}
lemma max_list1_as_Max:
  shows "max_list1 xs = Max (set xs)"
proof (cases "xs = []")
  case True
  then show ?thesis by (auto simp: Max.eq_fold' option.the_def max_list1_def)
next
  case False
  then show ?thesis
  proof (induct xs)
    case (Cons x xs)
    then show ?case by (cases "xs = []", auto simp: max_def)
  qed auto
qed

(* TODO: Replace AFP/Matrix/Utility/max_list *)
definition max_list where
  "max_list xs \<equiv> foldr max xs bot"

lemma max_list_Nil[simp]: "max_list [] = bot"
  and max_list_Cons[simp]: "max_list (x#xs) = max x (max_list xs)"
  by (auto simp: max_list_def)

lemma (in order_bot) max_bot [simp]: (* no need for order *)
  "max bot x = x" "max x bot = x" by (auto simp: max_def)

lemma max_list_as_max_list1_bot:
  shows "max_list xs = max_list1 (xs @ [bot])"
  by (induct xs, auto)

lemma max_list_as_max_list1:
  fixes xs :: "'a :: order_bot list"
  assumes "xs \<noteq> []" shows "max_list xs = max_list1 xs"
  by (insert assms, induct xs, auto)


lemma max_list_append [simp]:
  fixes xs ys :: "'a :: {linorder,order_bot} list"
  shows "max_list (xs @ ys) = max (max_list xs) (max_list ys)"
  by (induct xs arbitrary: ys, auto simp: ac_simps)

lemma max_list_concat:
  fixes xss :: "'a :: {linorder,order_bot} list list"
  shows "max_list (concat xss) = max_list (map max_list xss)"
  by (induct xss, auto)

lemma max_list1_append:
  fixes xs ys :: "'a :: linorder list"
  shows "max_list1 (xs @ ys) = (if xs = [] then max_list1 ys else if ys = [] then max_list1 xs else max (max_list1 xs) (max_list1 ys))"
  by (induct xs arbitrary: ys, auto simp: ac_simps)

lemma max_list1_concat:
  fixes xss :: "'a :: linorder list list"
  assumes "[] \<notin> set xss"
  shows "max_list1 (concat xss) = max_list1 (map max_list1 xss)"
  by (insert assms, induct xss, simp, insert List.last_in_set, force simp: max_list1_append)

lemma max_list1_mem: "xs \<noteq> [] \<Longrightarrow> max_list1 xs \<in> set xs"
proof (induct xs)
  case (Cons x xs)
  then show ?case by (cases xs, auto simp: max_def)
qed auto

lemma max_list_map_filter:
  fixes f :: "_ \<Rightarrow> 'a :: order_bot"
  assumes "\<And>x. x \<in> set xs \<Longrightarrow> \<not> P x \<Longrightarrow> f x = bot"
  shows "max_list (map f (filter P xs)) = max_list (map f xs)"
  using assms by (induct xs, auto)

lemma max_list_cong:
  assumes "\<And>x. x \<in> set xs \<Longrightarrow> f x = g x"
  shows "max_list (map f xs) = max_list (map g xs)"
  using assms by (induct xs, auto)

lemma hom_max:
  assumes "f x = f y \<or> (f x \<le> f y \<longleftrightarrow> x \<le> y)"
  shows "f (max x y) = max (f x) (f y)"
  using assms by (auto simp: max_def)

lemma hom_max_list1:
  assumes "\<And>x y. x \<in> set xs \<Longrightarrow> y \<in> set xs \<Longrightarrow> f x = f y \<or> (f x \<le> f y \<longleftrightarrow> x \<le> y)" and "xs \<noteq> []"
  shows "f (max_list1 xs) = max_list1 (map f xs)"
proof (insert assms, induct xs)
  case IH: (Cons x xs)
  show ?case
  proof (cases xs)
    case [simp]: (Cons y ys)
    from IH have "max_list1 (map f xs) = f (max_list1 xs)" by auto
    with max_list1_mem[of xs] hom_max[of f, OF IH.prems(1)]
    show ?thesis by auto
  qed auto
qed auto

(* should this really be called ord_hom? *)
locale ord_hom = fixes hom
  assumes ord_hom_condition: "\<And>x y. hom x = hom y \<or> (hom x \<le> hom y \<longleftrightarrow> x \<le> y)"
begin
lemmas hom_max = hom_max[of hom, OF ord_hom_condition]
lemmas hom_max_list1 = hom_max_list1[of _ hom, OF ord_hom_condition]
end

locale ord_hom_strict = fixes hom
  assumes hom_le_hom[simp]: "\<And>x y. hom x \<le> hom y \<longleftrightarrow> x \<le> y"
begin
sublocale ord_hom by (unfold_locales, auto)
end

locale linorder_hom = fixes hom :: "'a :: linorder \<Rightarrow> 'b :: linorder"
  assumes mono: "x \<le> y \<Longrightarrow> hom x \<le> hom y"
begin

sublocale ord_hom using mono by (unfold_locales, auto)

text {* @{thm mono_Max_commute} is derivable *}
lemma hom_Max:
  assumes finX: "finite X" and X0: "X \<noteq> {}" shows "hom (Max X) = Max (hom ` X)"
proof-
  obtain xs where "set xs = X" using finX finite_list by auto
  with X0 hom_max_list1[unfolded max_list1_as_Max, of xs] show ?thesis by auto
qed

end

interpretation add_hom: ord_hom "\<lambda>y :: 'a :: ordered_ab_semigroup_add_imp_le. x + y"
  by (unfold_locales, auto)

interpretation add_hom_right: ord_hom "\<lambda>x :: 'a :: ordered_ab_semigroup_add_imp_le. x + y"
  by (unfold_locales, auto)

interpretation int_hom: linorder_hom int by (unfold_locales, auto)

lemma sum_list_max_list1:
  fixes xss :: "'a :: {linordered_semiring} list list"
  assumes "[] \<notin> set xss"
  shows "sum_list (map max_list1 xss) = max_list1 (map sum_list (product_lists xss))"
proof (insert assms, induct xss)
  case (Cons xs xss)
  show ?case
  proof (cases "xss = []")
    case False
    have [simp]: "max_list1 (map (\<lambda>xs. x + sum_list xs) (product_lists xss)) = x + sum_list (map max_list1 xss)" for x
    proof-
      have "map (\<lambda>xs. x + sum_list xs) (product_lists xss) = map (\<lambda>y. x + y) (map sum_list (product_lists xss))"
        by auto
      also have "max_list1 \<dots> = x + max_list1 (map sum_list (product_lists xss))"
        apply (rule hom_max_list1[symmetric])
        using False Cons by auto
      also have "\<dots> = x + sum_list (map max_list1 xss)" using Cons by auto
      finally show ?thesis.
    qed
    show ?thesis
      apply (auto simp:map_concat)
      apply (subst max_list1_concat)
      using Cons apply force
      apply (simp add: o_def)
      apply (subst add_hom_right.hom_max_list1) using False Cons.prems by auto
  qed (auto simp: o_def)
qed simp

lemma prod_list_max_list1:
  fixes xss :: "'a :: {linordered_nonzero_semiring,linordered_semiring} list list"
  assumes "[] \<notin> set xss" and "\<And>x. x \<in> \<Union>(set ` set xss) \<Longrightarrow> x \<ge> 0"
  shows "prod_list (map max_list1 xss) = max_list1 (map prod_list (product_lists xss))"
proof (insert assms, induct xss)
  case (Cons xs xss)
  show ?case
  proof (cases "xss = []")
    case False
    { fix x :: 'a assume "x \<in> set xs"
      with Cons.prems have x0: "x \<ge> 0" by auto
      have "map (\<lambda>xs. x * prod_list xs) (product_lists xss) = map (\<lambda>y. x * y) (map prod_list (product_lists xss))"
        by auto
      also have "max_list1 \<dots> = x * max_list1 (map prod_list (product_lists xss))"
        apply (rule hom_max_list1[symmetric])
        using False Cons.prems x0 by (auto intro!: monoI mult_left_mono)
      also have "\<dots> = x * prod_list (map max_list1 xss)" using Cons by auto
      finally have "max_list1 (map (\<lambda>xs. x * prod_list xs) (product_lists xss)) = \<dots>".
    }
    note 1 = map_cong[OF refl this]
    from Cons.prems have "(\<And>x. x \<in> \<Union>(set ` set xss) \<Longrightarrow> x \<ge> 0)" "[] \<notin> set xss" by auto
    then have 2: "prod_list (map max_list1 xss) \<ge> 0"
    proof (induct xss)
      case (Cons xs xss)
      then have "max_list1 xs \<ge> 0" using max_list1_mem[of xs] by auto
      with Cons show ?case by auto
    qed (auto)
    from Cons.prems have 3: "xs \<noteq> []" by auto
    have "max_list1 xs \<ge> 0" using Cons.prems max_list1_mem[of xs] by auto
    show ?thesis
      apply (auto simp:map_concat)
      apply (subst max_list1_concat)
      using Cons.prems apply force
      apply (simp add: o_def)
      apply (subst 1, assumption)
      apply (subst hom_max_list1[of _ "\<lambda>x. x * _"]) using mult_right_mono[OF _2] 3 by auto
  qed (auto simp: o_def)
qed simp


text {* We implement max-polynomial interpretations as an instance of @{locale encoded_algebra}.
  As a target algebra, we introduce a $G$-algebra where $G = \{+,\times,\max\}$. *}

locale max_poly_locale
begin

datatype sig = ConstF nat | SumF | ProdF | MaxF

primrec I where
  "I (ConstF n) = (\<lambda>x. n)"
| "I SumF = sum_list"
| "I ProdF = prod_list"
| "I MaxF = max_list"

abbreviation const where "const n \<equiv> Fun (ConstF n) []"

sublocale wf_weak_mono_algebra less_eq less I
proof
  fix f ls rs and a b :: nat
  assume ba: "b \<le> a"
  then show "I f (ls @ b # rs) \<le> I f (ls @ a # rs)"
  proof(cases f)
    case [simp]: MaxF
    note ba
    also have "a \<le> max_list (ls @ a # rs)" by auto
    finally show ?thesis by auto
  qed auto
qed

sublocale algebra_constant I const by (unfold_locales, auto)

fun simplify where
  "simplify (Var x) = Var x"
| "simplify (Fun (ConstF n) ss) = const n"
| "simplify (Fun SumF ss) = (
    let ss' = filter (\<lambda>s. s \<noteq> const 0) (map simplify ss) in
    case ss' of [] \<Rightarrow> const 0 | [s] \<Rightarrow> s | _ \<Rightarrow> Fun SumF ss'
  )"
| "simplify (Fun ProdF ss) = (
    let ss' = filter (\<lambda>s. s \<noteq> const 1) (map simplify ss) in
    if const 0 \<in> set ss' then const 0
    else case ss' of [] \<Rightarrow> const 1 | [s] \<Rightarrow> s | _ \<Rightarrow> Fun ProdF ss'
  )"
| "simplify (Fun MaxF ss) = (
    let ss' = filter (\<lambda>s. s \<noteq> const 0) (map simplify ss) in
    case ss' of [] \<Rightarrow> const 0 | [s] \<Rightarrow> s | _ \<Rightarrow> Fun MaxF ss'
  )"

lemma list_caseE:
  assumes "case xs of [] \<Rightarrow> a | x # ys \<Rightarrow> b x ys"
    and "xs = [] \<Longrightarrow> a \<Longrightarrow> thesis"
    and "\<And>x ys. xs = x#ys \<Longrightarrow> b x ys \<Longrightarrow> thesis"
  shows thesis
  using assms
  by (cases xs, auto)

lemma vars_term_simplify: "vars_term (simplify s) \<subseteq> vars_term s"
proof (rule subsetI, induct s rule: simplify.induct)
  case (1 v)
  then show ?case by simp
next
  case (2 n ss)
  then show ?case by simp
next
  case IH: (3 ss)
  from IH.prems have "x \<in> \<Union>(vars_term ` set (filter (\<lambda>s. s \<noteq> const 0) (map simplify ss)))"
    by (auto simp: Let_def list.case_distrib simp del: set_filter elim!: list_caseE)
  also have "\<dots> \<subseteq> vars_term (Fun SumF ss)" using IH.hyps by auto
  finally show ?case.
next
  case IH: (4 ss)
  let ?ss = "filter (\<lambda>s. s \<noteq> const 1) (map simplify ss)"
  from IH.prems have "x \<in> \<Union>(vars_term ` set ?ss)"
    by (cases "const 0 \<in> set ?ss", auto simp: Let_def list.case_distrib simp del: set_filter elim!: list_caseE)
  also have "\<dots> \<subseteq> vars_term (Fun ProdF ss)" using IH.hyps by auto
  finally show ?case.
next
  case IH: (5 ss)
  from IH.prems have "x \<in> \<Union>(vars_term ` set (filter (\<lambda>s. s \<noteq> const 0) (map simplify ss)))"
    by (auto simp: Let_def list.case_distrib simp del: set_filter elim!: list_caseE)
  also have "\<dots> \<subseteq> vars_term (Fun MaxF ss)" using IH.hyps by auto
  finally show ?case.
qed



lemma eval_SumF_filter_0:
  "\<lbrakk>Fun SumF (filter (\<lambda>s. s \<noteq> const 0) ss)\<rbrakk> = \<lbrakk>Fun SumF ss\<rbrakk>"
  by (auto intro!: ext sum_list_map_filter)

lemma eval_MaxF_filter_0:
  "\<lbrakk>Fun MaxF (filter (\<lambda>s. s \<noteq> const 0) ss)\<rbrakk> = \<lbrakk>Fun MaxF ss\<rbrakk>"
  by (auto intro!: ext max_list_map_filter)

lemma eval_ProdF_filter_1:
  "\<lbrakk>Fun ProdF (filter (\<lambda>s. s \<noteq> const (Suc 0)) ss)\<rbrakk> = \<lbrakk>Fun ProdF ss\<rbrakk>"
  by (auto intro!: ext prod_list_map_filter)

lemma eval_ProdF_with_0:
  "\<exists>s \<in> set ss. s = const 0 \<Longrightarrow> \<lbrakk>Fun ProdF ss\<rbrakk> = \<lbrakk>const 0\<rbrakk>"
  by (intro ext, induct ss, auto)

lemma simplify_lemma:
  "\<lbrakk>case ss of [] \<Rightarrow> const 0 | [s] \<Rightarrow> s | _ \<Rightarrow> Fun SumF ss\<rbrakk> = \<lbrakk>Fun SumF ss\<rbrakk>"
  "\<lbrakk>case ss of [] \<Rightarrow> const 1 | [s] \<Rightarrow> s | _ \<Rightarrow> Fun ProdF ss\<rbrakk> = \<lbrakk>Fun ProdF ss\<rbrakk>"
  "\<lbrakk>case ss of [] \<Rightarrow> const 0 | [s] \<Rightarrow> s | _ \<Rightarrow> Fun MaxF ss\<rbrakk> = \<lbrakk>Fun MaxF ss\<rbrakk>"
  by (atomize(full), cases ss rule: list_3_cases, auto simp: bot_nat_def)

lemma simplify [simp]: "\<lbrakk>simplify s\<rbrakk> = \<lbrakk>s\<rbrakk>"
proof (intro ext, induct s rule: simplify.induct)
  case (1 x \<alpha>)
  then show ?case by simp
next
  case (2 n ss \<alpha>)
  then show ?case by (auto simp: simplify_lemma)
next
  case (3 ss \<alpha>)
  then show ?case by (auto simp: Let_def simplify_lemma eval_SumF_filter_0 intro!:sum_list_cong)
next
  case (4 ss \<alpha>)
  then show ?case
    by (force simp: Let_def simplify_lemma eval_ProdF_filter_1[simplified] prod_list_zero_iff image_def intro: prod_list_cong)
next
  case (5 ss \<alpha>)
  then show ?case by (auto simp: Let_def simplify_lemma eval_MaxF_filter_0 o_def intro:max_list_cong)
qed

end

derive compare max_poly_locale.sig
thm weak_mono_algebra.weak_mono
interpretation max_poly: max_poly_locale.


subsection {* Translation to Integer Arithmetic *}

context IA_locale
begin

lemma to_int_iff:
  assumes \<alpha>: "assignment \<alpha>" and s: "s :\<^sub>f IntT" and t: "t :\<^sub>f IntT"
  shows "to_int (\<lbrakk>s\<rbrakk>\<alpha>) < to_int (\<lbrakk>t\<rbrakk>\<alpha>) \<longleftrightarrow> \<lbrakk>s\<rbrakk>\<alpha> < \<lbrakk>t\<rbrakk>\<alpha>"
    and "to_int (\<lbrakk>s\<rbrakk>\<alpha>) \<le> to_int (\<lbrakk>t\<rbrakk>\<alpha>) \<longleftrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<le> \<lbrakk>t\<rbrakk>\<alpha>"
    and "to_int (\<lbrakk>s\<rbrakk>\<alpha>) = to_int (\<lbrakk>t\<rbrakk>\<alpha>) \<longleftrightarrow> \<lbrakk>s\<rbrakk>\<alpha> = \<lbrakk>t\<rbrakk>\<alpha>"
  using s t by (auto dest!: eval_types[OF \<alpha>])

end

context max_poly_locale
begin

primrec to_IA where
  "to_IA (Var x) = [Var (x,IA.IntT)]"
| "to_IA (Fun f ss) = (case f of (ConstF n) \<Rightarrow> [Fun (IA.ConstF n) []]
    | MaxF \<Rightarrow> if ss = [] then [IA.const 0] else concat (map to_IA ss)
    | SumF \<Rightarrow>
      if ss = [] then [Fun (IA.SumF 0) []]
      else map (Fun (IA.SumF (length ss))) (product_lists (map to_IA ss))
    | ProdF \<Rightarrow>
      if ss = [] then [Fun (IA.ProdF 0) []]
      else map (Fun (IA.ProdF (length ss))) (product_lists (map to_IA ss))
  )"

lemma to_IA_induct[case_names Var ConstF SumF ProdF MaxF SumF0 ProdF0 MaxF0]:
  assumes "\<And>x. P (Var x)"
    and "\<And>n ss. P (Fun (ConstF n) ss)"
    and "\<And>ss. ss \<noteq> [] \<Longrightarrow> (\<And>s. s \<in> set ss \<Longrightarrow> P s) \<Longrightarrow> P (Fun SumF ss)"
    and "\<And>ss. ss \<noteq> [] \<Longrightarrow> (\<And>s. s \<in> set ss \<Longrightarrow> P s) \<Longrightarrow> P (Fun ProdF ss)"
    and "\<And>ss. ss \<noteq> [] \<Longrightarrow> (\<And>s. s \<in> set ss \<Longrightarrow> P s) \<Longrightarrow> P (Fun MaxF ss)"
    and "P (Fun SumF [])"
    and "P (Fun ProdF [])"
    and "P (Fun MaxF [])"
  shows "P s"
  by (insert assms, induction_schema, metis term.exhaust sig.exhaust, lexicographic_order)


lemma to_IA_has_type:
  "s' \<in> set (to_IA s) \<Longrightarrow> IA.has_type s' IA.IntT"
  apply (induct s arbitrary:s' rule: to_IA_induct, auto simp:in_set_product_lists_length product_lists_set)
  apply (metis length_map list_all2_nthD2 nth_map nth_mem)
  apply (metis length_map list_all2_nthD2 nth_map nth_mem)
  done

lemma to_IA_eq_NilE[elim!]: "to_IA s = [] \<Longrightarrow> thesis" by (induct s rule: to_IA_induct, auto)

definition "to_IA_assignment \<alpha> \<equiv> \<lambda>(x,t). case t of IA.IntT \<Rightarrow> IA.Int (int (\<alpha> x)) | _ \<Rightarrow> IA.Bool False"

lemma to_IA_assignment: "IA.assignment (to_IA_assignment \<alpha>)"
  by (intro IA.assignmentI, auto simp: to_IA_assignment_def split: IA.ty.split)

definition "eval_via_IA s \<alpha> \<equiv> map (\<lambda>t. IA.eval t (to_IA_assignment \<alpha>)) (to_IA s)"

lemma eval_via_IA_eq_NilE[elim!]: "eval_via_IA s \<alpha> = [] \<Longrightarrow> thesis"
  by (auto simp: eval_via_IA_def)

lemma eval_via_IA_range: "set (eval_via_IA s \<alpha>) \<subseteq> range IA.Int"
proof-
  have "set (eval_via_IA s \<alpha>) \<subseteq> set (map (\<lambda>t. IA.eval t (to_IA_assignment \<alpha>)) (to_IA s) @ [IA.Int 0])"
    by (unfold eval_via_IA_def, auto)
  also have "\<dots> \<subseteq> range IA.Int"
    using to_IA_assignment[of \<alpha>]
    by (auto dest!: to_IA_has_type IA.eval_types)
  finally show ?thesis.
qed

lemma max_list1_eval_via_IA_range:
  "max_list1 (eval_via_IA s \<alpha>) \<in> range IA.Int"
proof-
  note max_list1_mem
  also note eval_via_IA_range
  finally show ?thesis by force
qed

end

(* TODO: Move *)
interpretation int_hom: comm_semiring_hom int by (unfold_locales, auto)

instantiation IA.val :: plus
begin
fun plus_val where
  "plus_val (IA.Int x) (IA.Int y) = IA.Int (x + y)"
| "plus_val (IA.Bool x) (IA.Bool y) = IA.Bool (x \<or> y)"
instance..
end

instantiation IA.val :: times
begin
fun times_val where
  "times_val (IA.Int x) (IA.Int y) = IA.Int (x * y)"
| "times_val (IA.Bool x) (IA.Bool y) = IA.Bool (x \<and> y)"
instance..
end


definition listprod where "listprod G xs \<equiv> foldr (mult G) xs (one G)"

lemma listprod_Nil[simp]: "listprod G [] = one G" by (auto simp: listprod_def)

lemma listprod_Cons[simp]: "listprod G (x#xs) = mult G x (listprod G xs)" by (auto simp: listprod_def)

lemma nth_ConsI: "(i = 0 \<Longrightarrow> P x) \<Longrightarrow> (\<And>j. i = Suc j \<Longrightarrow> P (xs!j)) \<Longrightarrow> P ((x#xs)!i)"
  by (cases i, auto)

lemma (in comm_monoid) listprod_as_finprod:
  assumes "set xs \<subseteq> carrier G"
  shows "listprod G xs = finprod G (nth xs) {..<length xs}"
  apply (insert assms, induct xs, auto simp: lessThan_Suc_eq_insert_0)
  apply (subst finprod_insert)
  apply (auto intro!: nth_ConsI[of _ "\<lambda>x. x \<in> carrier G"])
  apply (subst finprod_reindex, auto)
  done

definition listsum where "listsum R \<equiv> listprod \<lparr> carrier = carrier R, mult = add R, one = zero R \<rparr>"

lemma listsum_Nil[simp]: "listsum R [] = zero R" by (auto simp: listsum_def)

lemma listsum_Cons[simp]: "listsum R (x#xs) = add R x (listsum R xs)" by (auto simp: listsum_def)

lemmas (in ring) listsum_as_finsum = add.listprod_as_finprod[folded listsum_def]


context IA_locale begin

definition "ring_Int \<equiv> \<lparr> carrier = range IA.Int, mult = op *, one = IA.Int 1, zero = IA.Int 0, add = op + \<rparr>"

end

interpretation IA_ring_Int: Ring.ring "IA.ring_Int"
  rewrites [simp]: "carrier IA.ring_Int = range IA.Int"
    and [simp]: "mult IA.ring_Int = op *"
    and [simp]: "one IA.ring_Int = IA.Int 1"
    and [simp]: "zero IA.ring_Int = IA.Int 0"
    and [simp]: "add IA.ring_Int = op +"
    and [simp]: "Units IA.ring_Int = { IA.Int 1, IA.Int (-1) }"
  by (unfold_locales, auto simp: IA.ring_Int_def Units_def field_simps zmult_eq_1_iff intro: exI[of _ "-_"])

interpretation IA_Int_hom: plus_hom IA.Int by (unfold_locales, auto)

interpretation IA_Int_hom: ord_hom_strict IA.Int by (unfold_locales, auto)

lemma IA_Int_sum_list: "IA.Int (sum_list xs) = listsum IA.ring_Int (map IA.Int xs)"
  by (induct xs, auto simp:lessThan_Suc_eq_insert_0 IA_Int_hom.hom_add)

lemma listsum_IA:
  assumes "set vs \<subseteq> range IA.Int"
  shows "listsum IA.ring_Int vs = IA.Int (sum_list (map IA.to_int vs))"
  using IA_Int_sum_list[of "map IA.to_int vs"] assms map_idI[of vs "IA.Int \<circ> IA.to_int"] by force

context max_poly_locale
begin

lemma eval_via_IA_ge_0: "v \<in> set (eval_via_IA s \<alpha>) \<Longrightarrow> IA.to_int v \<ge> 0"
proof (unfold eval_via_IA_def, induct s arbitrary:v rule: to_IA_induct)
  case (Var x)
  then show ?case by (auto simp: to_IA_assignment_def)
next
  case (SumF ss)
  from this(1,3)
  show ?case
    apply (auto simp: o_def image_def in_set_product_lists_length intro!:sum_list_ge_0_nth)
    apply(drule(1) in_set_product_lists_nth[of _ "map to_IA ss", simplified])
    using SumF(2)[OF nth_mem] apply auto
    done
next
  case (ProdF ss)
  from this(1,3)
  show ?case
    apply (auto simp: o_def image_def in_set_product_lists_length intro!:prod_list_nonneg_nth)
    apply(drule  in_set_product_lists_nth[of _ "map to_IA ss", simplified])
    using ProdF(2)[OF nth_mem] apply auto
    done
qed auto

lemma eval_via_IA:
  fixes f \<alpha>
  defines "f \<equiv> \<lambda>s. eval_via_IA s \<alpha>"
  shows "int (\<lbrakk>s\<rbrakk> \<alpha>) = IA.to_int (max_list1 (f s))"
proof (induct s rule: to_IA_induct)
  case (Var x)
  show ?case by (simp add: eval_via_IA_def to_IA_assignment_def IA_Int_hom.hom_max[symmetric] f_def)
next
  case (ConstF n ss)
  show ?case by (simp add: eval_via_IA_def IA_Int_hom.hom_max[symmetric] f_def)
next
  case (SumF ss)
  note ss = this(1) and 1 = map_ext[OF this(2)[unfolded atomize_imp]]
  have "int (\<lbrakk>Fun SumF ss\<rbrakk> \<alpha>) = (\<Sum>s\<leftarrow>ss. IA.to_int (max_list1 (f s)))"
    by (auto simp: image_def int_hom.hom_sum_list o_def 1)
  also have "\<dots> = (\<Sum>s\<leftarrow>ss. (max_list1 (map IA.to_int (f s))))"
    apply (rule arg_cong[of _ _sum_list])
    apply (rule map_cong[OF refl])
    apply (rule hom_max_list1)
    by (auto simp: o_def f_def dest!: subsetD[OF eval_via_IA_range])
  also have "\<dots> = sum_list (map max_list1 (map (map IA.to_int) (map f ss)))"
    by (auto simp: o_def)
  also have "\<dots> = max_list1 (map sum_list (product_lists (map (map IA.to_int) (map f ss))))"
    by (intro sum_list_max_list1, auto simp:o_def f_def)
  also have "\<dots> = max_list1 (map sum_list (product_lists (map (map (IA.to_int \<circ> (\<lambda>t. IA.eval t (to_IA_assignment \<alpha>)))) (map to_IA ss))))"
    by (auto simp: o_def f_def eval_via_IA_def)
  also have "\<dots> = max_list1 (map IA.to_int (f (Fun SumF ss)))"
    by (subst product_lists_map_map, auto simp: o_def f_def eval_via_IA_def ss)
  also have "\<dots> = IA.to_int (max_list1 (f (Fun SumF ss)))"
    by (rule hom_max_list1[symmetric], auto simp: f_def dest!: subsetD[OF eval_via_IA_range])
  finally show ?case.
next
  case (ProdF ss)
  note ss = this(1) and 1 = map_ext[OF this(2)[unfolded atomize_imp]]
  have "int (\<lbrakk>Fun ProdF ss\<rbrakk> \<alpha>) = (\<Prod>s\<leftarrow>ss. IA.to_int (max_list1 (f s)))"
    by (auto simp: image_def int_hom.hom_prod_list o_def 1)
  also have "\<dots> = (\<Prod>s\<leftarrow>ss. (max_list1 (map IA.to_int (f s))))"
    apply (rule arg_cong[of _ _prod_list])
    apply (rule map_cong[OF refl])
    apply (rule hom_max_list1)
    by (auto simp: o_def f_def dest!: subsetD[OF eval_via_IA_range])
  also have "\<dots> = prod_list (map max_list1 (map (map IA.to_int) (map f ss)))"
    by (auto simp: o_def)
  also have "\<dots> = max_list1 (map prod_list (product_lists (map (map IA.to_int) (map f ss))))"
    by (intro prod_list_max_list1, auto simp:o_def f_def intro:eval_via_IA_ge_0)
  also have "\<dots> = max_list1 (map prod_list (product_lists (map (map (IA.to_int \<circ> (\<lambda>t. IA.eval t (to_IA_assignment \<alpha>)))) (map to_IA ss))))"
    by (auto simp: o_def f_def eval_via_IA_def)
  also have "\<dots> = max_list1 (map IA.to_int (f (Fun ProdF ss)))"
    by (subst product_lists_map_map, auto simp: o_def f_def eval_via_IA_def ss)
  also have "\<dots> = IA.to_int (max_list1 (f (Fun ProdF ss)))"
    by (rule hom_max_list1[symmetric], auto simp: f_def dest!: subsetD[OF eval_via_IA_range])
  finally show ?case.
next
  case (MaxF ss)
  note ss = this(1) and 1 = map_ext[OF this(2)[unfolded atomize_imp]]
  have "\<lbrakk>Fun MaxF ss\<rbrakk>\<alpha> = max_list1 (map (\<lambda>s. \<lbrakk>s\<rbrakk>\<alpha>) ss)"
    using ss by (auto simp: max_list_as_max_list1)
  also have "int \<dots> = max_list1 (map (\<lambda>x. IA.to_int (max_list1 (f x))) ss)"
    using ss by (auto simp: o_def int_hom.hom_max_list1 1)
  also have "\<dots> = max_list1 (map (\<lambda>s. max_list1 (map IA.to_int (f s))) ss)"
    apply (rule arg_cong[of _ _ max_list1])
    apply (rule map_cong[OF refl])
    apply (rule hom_max_list1)
    by (auto simp: o_def f_def dest!: subsetD[OF eval_via_IA_range])
  also have "\<dots> = max_list1 (map max_list1 (map (map IA.to_int) (map f ss)))"
    by (auto simp: o_def)
  also have "\<dots> = max_list1 (concat (map (map IA.to_int \<circ> f) ss))"
    apply (subst max_list1_concat[symmetric])
    using ss by (auto simp: f_def eval_via_IA_def)
  also have "\<dots> = max_list1 (concat (map (map (IA.to_int \<circ> (\<lambda>t. IA.eval t (to_IA_assignment \<alpha>)))) (map to_IA ss)))"
    by (auto simp: o_def f_def eval_via_IA_def)
  also have "\<dots> = max_list1 (map IA.to_int (f (Fun MaxF ss)))"
    apply (unfold concat_map_map)
    by (auto simp: o_def f_def eval_via_IA_def ss)
  also have "\<dots> = IA.to_int (max_list1 (f (Fun MaxF ss)))"
    by (rule hom_max_list1[symmetric], auto simp: f_def dest!: subsetD[OF eval_via_IA_range])
  finally show ?case.
qed (auto simp: f_def eval_via_IA_def bot_nat_def)

definition le_via_IA (infix "\<sqsubseteq>\<^sub>I" 50) where "s \<sqsubseteq>\<^sub>I t \<equiv>
  Conjunction (map (\<lambda>s'. disj_atoms (map (\<lambda>t'. Fun IA.LeF [s', t']) (to_IA t))) (to_IA s))"

lemma formula_le_via_IA [intro!]: "IA.formula (s \<sqsubseteq>\<^sub>I t)"
  by (auto simp: le_via_IA_def less_Suc_eq dest: to_IA_has_type)

lemma le_via_IA:
  assumes "IA.valid (s \<sqsubseteq>\<^sub>I t)" shows "s \<sqsubseteq>\<^sub>t t"
proof
  fix \<alpha> :: "_ \<Rightarrow> nat"
  from IA.validD[OF assms to_IA_assignment, of \<alpha>]
  have "\<forall>s' \<in> set (to_IA s). \<exists>t' \<in> set (to_IA t).
    IA.eval s' (to_IA_assignment \<alpha>) \<le> IA.eval t' (to_IA_assignment \<alpha>)"
    by (auto simp: le_via_IA_def IA.to_int_iff[OF to_IA_assignment to_IA_has_type to_IA_has_type] elim!: ballE)
  then have "max_list1 (eval_via_IA s \<alpha>) \<le> max_list1 (eval_via_IA t \<alpha>)"
    by (unfold max_list1_as_Max eval_via_IA_def, subst Max_le_Max, auto)
  then have "int (\<lbrakk>s\<rbrakk> \<alpha>) \<le> int (\<lbrakk>t\<rbrakk> \<alpha>)"
    unfolding eval_via_IA
    by (metis IA_Int_hom.hom_le_hom IA.val.sel(1) imageE max_list1_eval_via_IA_range)
  then show "\<lbrakk>s\<rbrakk> \<alpha> \<le> \<lbrakk>t\<rbrakk> \<alpha>" by auto
qed

definition less_via_IA (infix "\<sqsubset>\<^sub>I" 50) where "s \<sqsubset>\<^sub>I t \<equiv>
  Conjunction (map (\<lambda>s'. disj_atoms (map (\<lambda>t'. Fun IA.LessF [s', t']) (to_IA t))) (to_IA s))"

lemma formula_less_via_IA [intro!]: "IA.formula (s \<sqsubset>\<^sub>I t)"
  by (auto simp: less_via_IA_def less_Suc_eq dest: to_IA_has_type)


lemma less_via_IA:
  assumes "IA.valid (s \<sqsubset>\<^sub>I t)" shows "s \<sqsubset>\<^sub>t t"
proof
  fix \<alpha> :: "_ \<Rightarrow> nat"
  from IA.validD[OF assms to_IA_assignment, of \<alpha>]
  have "\<forall>s' \<in> set (to_IA s). \<exists>t' \<in> set (to_IA t). IA.eval s' (to_IA_assignment \<alpha>) < IA.eval t' (to_IA_assignment \<alpha>)"
    by (auto simp: less_via_IA_def IA.to_int_iff[OF to_IA_assignment to_IA_has_type to_IA_has_type] elim!: ballE)
  then have "max_list1 (eval_via_IA s \<alpha>) < max_list1 (eval_via_IA t \<alpha>)"
    by (unfold max_list1_as_Max eval_via_IA_def, subst Max_less_Max, auto)
  then have "int (\<lbrakk>s\<rbrakk> \<alpha>) < int (\<lbrakk>t\<rbrakk> \<alpha>)"
    unfolding eval_via_IA
    by (metis IA.less_IA_value_simps(3) IA.val.collapse(1) IA.val.discI(1) imageE max_list1_eval_via_IA_range)
  then show "\<lbrakk>s\<rbrakk> \<alpha> < \<lbrakk>t\<rbrakk> \<alpha>" by auto
qed

end


subsection {* Max-Polynomial Interpretations *}

text {* Now we encode arbitrary $F$-terms into the @{type max_poly.sig}-terms. *}

locale max_poly_encoding = encoded_ordered_algebra
  where less_eq = "op \<le>" and less = "op <" and I = max_poly.I
begin

sublocale encoded_wf_weak_mono_algebra where less_eq = "op \<le>" and less = "op <" and I = max_poly.I
  by (unfold_locales, fact max_poly.append_Cons_le_append_Cons)

lemmas redpair = redpair.redpair_axioms

lemma less_term_via_IA:
  assumes "IA.valid (encode s \<sqsubset>\<^sub>I encode t)" shows "s \<sqsubset>\<^sub>t t"
  using max_poly.less_via_IA[OF assms] by (auto intro!: less_termI)

lemma less_eq_term_via_IA:
  assumes "IA.valid (encode s \<sqsubseteq>\<^sub>I encode t)" shows "s \<sqsubseteq>\<^sub>t t"
  using max_poly.le_via_IA[OF assms] by (auto intro!: less_eq_termI)

lemma ws_fun_arg_via_IA: (* TODO: one can syntactically check ws_fun_arg. *)
  assumes "IA.valid (Var i \<sqsubseteq>\<^sub>I E f n)"
  shows "ws_fun_arg (rel_of op \<sqsupseteq>\<^sub>t) (f,n) i"
proof (intro ws_fun_argI, unfold mem_Collect_eq prod.case, intro less_eq_termI)
  fix ts :: "('a,'b) term list" and \<alpha>
  assume n: "length ts = n" and i_n: "i < n"
  note max_poly.le_via_IA[OF assms]
  then
  have "\<alpha> i \<le> max_poly.eval (E f n) \<alpha>" for \<alpha> by auto
  note this[of "\<lambda>i. \<lbrakk>ts!i\<rbrakk>\<alpha>"]
  also have "target.eval (E f n) (\<lambda>i. \<lbrakk>ts ! i\<rbrakk> \<alpha>) = \<lbrakk>Fun f ts\<rbrakk> \<alpha>"
    using n encoder.var_domain by auto
  finally show "\<lbrakk>ts ! i\<rbrakk> \<alpha> \<le> \<lbrakk>Fun f ts\<rbrakk> \<alpha>".
qed




end

end

